(******************************************************************************)
(** * Definition of execution graphs   *)
(******************************************************************************)

Require Import Classical Peano_dec.
Require Import Hahn.
Require Import AuxRel.
Require Import Labels Events.

Set Implicit Arguments.

(******************************************************************************)
(** Definition of an execution *)
(******************************************************************************)

Record execution :=
  { acts : list Event;
    rf : Event -> Event -> Prop ;
    co : Event -> Event -> Prop ;
  }.

Section Execution.

Variable G : execution.

Definition acts_set := fun x => In x G.(acts).

Notation "'E'" := acts_set.
Notation "'rf'" := G.(rf).
Notation "'co'" := G.(co).

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Definition sb := ⦗E⦘ ⨾ ext_sb ⨾  ⦗E⦘.

Record Wf := 
  { wf_index : ⦗E⦘ ⨾ same_tid ∩ same_index ⨾ ⦗E⦘ ⊆ (is_init × is_init) ∪ ⦗E⦘ ;
    wf_rfE : rf ≡ ⦗E⦘ ⨾ rf ⨾ ⦗E⦘ ;
    wf_rfD : rf ≡ ⦗W⦘ ⨾ rf ⨾ ⦗R⦘ ;
    wf_rfl : rf ⊆ same_loc ;
    wf_rfv : forall w r (RF: rf w r), valw w = valr r ;
    wf_rff : functional rf⁻¹ ;
    wf_coE : co ≡ ⦗E⦘ ⨾ co ⨾ ⦗E⦘ ;
    wf_coD : co ≡ ⦗W⦘ ⨾ co ⨾ ⦗W⦘ ;
    wf_col : co ⊆ same_loc ;
    co_trans : transitive co ;
    wf_co_total : forall x, is_total (E ∩₁ W ∩₁ (fun a => loc a = x)) co ;
    co_irr : irreflexive co ;
    wf_initE : is_init ⊆₁ E ;
    wf_co_init : co ≡ co ⨾  ⦗fun x => ~ is_init x⦘ ;
  }.

Implicit Type WF : Wf.

(******************************************************************************)
(** ** Derived relations  *)
(******************************************************************************)

(* reads-before, aka from-read *)
Definition fr := rf⁻¹ ⨾ co \ ⦗fun _ => True⦘.

(* happens-before *)
Definition hb := (sb ∪ rf)⁺.

(* happens-before-sc *)
Definition hb_sc := (sb ∪ rf ∪ co ∪ fr)⁺.

Definition complete := E ∩₁ R  ⊆₁ codom_rel rf.

(******************************************************************************)
(** ** Basic transitivity properties *)
(******************************************************************************)

Lemma sb_trans : transitive sb.
Proof.
unfold sb; unfolder; ins; desf; splits; auto.
eby eapply ext_sb_trans.
Qed.

Lemma sb_sb : sb ⨾ sb ⊆ sb.
Proof.
generalize sb_trans. basic_solver 21.
Qed.

Lemma sb_same_loc_trans: transitive (sb ∩ same_loc).
Proof.
apply transitiveI.
unfold Events.same_loc.
unfolder; ins; desf; eauto.
splits.
generalize sb_trans; basic_solver 21.
congruence.
Qed.

(******************************************************************************)
(** ** Basic properties *)
(******************************************************************************)

Lemma sb_neq_loc_in_sb : sb \ same_loc ⊆ sb.
Proof. basic_solver. Qed.

Lemma sb_in_hb : sb ⊆ hb.
Proof. unfold hb. rewrite <- ct_step. basic_solver. Qed.

Lemma hb_trans : transitive hb.
Proof. rels. Qed.

(******************************************************************************)
(** ** Same Location relations  *)
(******************************************************************************)

Lemma loceq_rf WF : funeq loc rf.
Proof. apply WF. Qed.
(* 
End Execution.

Print loceq_rf.
Print acts_set.
 


Definition aaa G (WF: Wf G) : funeq loc G.(rf) :=
(loceq_rf WF).

Definition bbb G (WF: Wf G) : G.(co) ⊆ same_loc :=
WF.(wf_col).

Definition ccc G (WF: Wf G) : G.(co) ⊆ same_loc :=
wf_col WF.

Notation "x ..( y )" := (y x) (at level 0, only parsing).

(* Arguments wf_col [ G ]. *)

Definition ddd G (WF: Wf G) : G.(co) ⊆ same_loc :=
WF..(wf_col).
*)

Lemma loceq_co WF : funeq loc co.
Proof. apply WF. Qed.

Lemma loceq_fr WF : funeq loc fr.
Proof.
unfold funeq.
unfold fr; unfolder; ins; desf.
generalize (loceq_co WF), (loceq_rf WF).
transitivity (loc z); [symmetry; eauto|eauto].
Qed.

Lemma wf_frl WF : fr ⊆ same_loc.
Proof.
unfold fr.
rewrite (wf_rfl WF), (wf_col WF).
unfold Events.same_loc.
unfolder; ins; desc; congruence. 
Qed.

(******************************************************************************)
(** ** Relations in graph *)
(******************************************************************************)

Lemma wf_sbE : sb ≡ ⦗E⦘ ⨾ sb ⨾ ⦗E⦘.
Proof. 
split; [|basic_solver].
unfold sb; basic_solver 42. 
Qed.

Lemma wf_frE WF : fr ≡ ⦗E⦘ ⨾ fr ⨾ ⦗E⦘.
Proof.
split; [|basic_solver].
unfold fr.
rewrite (wf_rfE WF) at 1.
rewrite (wf_coE WF) at 1.
basic_solver.
Qed.

Lemma wf_hbE WF : hb ≡ ⦗E⦘ ⨾ hb ⨾ ⦗E⦘.
Proof.
split; [|basic_solver].
unfold hb.
rewrite <- inclusion_ct_seq_eqv_r, <- inclusion_ct_seq_eqv_l.
apply inclusion_t_t.
rewrite wf_sbE at 1.
rewrite (wf_rfE WF) at 1.
basic_solver 42.
Qed.

Lemma wf_hb_scE WF : hb_sc ≡ ⦗E⦘ ⨾ hb_sc ⨾ ⦗E⦘.
Proof.
split; [|basic_solver].
unfold hb_sc.
rewrite <- inclusion_ct_seq_eqv_r, <- inclusion_ct_seq_eqv_l.
apply inclusion_t_t.
rewrite wf_sbE at 1.
rewrite (wf_rfE WF) at 1.
rewrite (wf_coE WF) at 1.
rewrite (wf_frE WF) at 1.
basic_solver 42.
Qed.

Lemma wf_ct_rfE WF : (rf)⁺ ≡ ⦗E⦘ ⨾ (rf)⁺ ⨾ ⦗E⦘.
Proof.
split; [|basic_solver].
rewrite <- inclusion_ct_seq_eqv_r, <- inclusion_ct_seq_eqv_l.
apply inclusion_t_t.
rewrite (wf_rfE WF) at 1.
basic_solver 42.
Qed.

Lemma wf_ct_transp_rfE WF : (rf⁻¹)⁺ ≡ ⦗E⦘ ⨾ (rf^{-1})⁺ ⨾ ⦗E⦘.
Proof.
split; [|basic_solver].
rewrite <- inclusion_ct_seq_eqv_r, <- inclusion_ct_seq_eqv_l.
apply inclusion_t_t.
rewrite (wf_rfE WF) at 1.
basic_solver 42.
Qed.


(******************************************************************************)
(** ** Domains and codomains  *)
(******************************************************************************)

Lemma wf_frD WF : fr ≡ ⦗R⦘ ⨾ fr ⨾ ⦗W⦘.
Proof.
split; [|basic_solver].
unfold fr.
rewrite (wf_rfD WF) at 1.
rewrite (wf_coD WF) at 1.
basic_solver.
Qed.

(******************************************************************************)
(** ** Irreflexive relations *)
(******************************************************************************)

Lemma sb_irr : irreflexive sb.
Proof.
unfold sb; unfolder; ins; desf.
eby eapply ext_sb_irr.
Qed.

Lemma fr_irr WF : irreflexive fr.
Proof.
rewrite wf_frE; unfold fr; basic_solver.
Qed.

(******************************************************************************)
(** ** Acyclic relations *)
(******************************************************************************)

Lemma sb_acyclic : acyclic sb.
Proof.
apply trans_irr_acyclic; [apply sb_irr| apply sb_trans]. 
Qed.

Lemma co_acyclic WF: acyclic co.
Proof.
by apply trans_irr_acyclic; [apply co_irr| apply co_trans]. 
Qed.

(******************************************************************************)
(** ** init *)
(******************************************************************************)

Lemma init_w WF: is_init ⊆₁ W \₁ RMW.
Proof.
unfolder; ins.
unfold is_init in *; destruct x; desf.
Qed.

Lemma read_or_rmw_not_init WF a (A: R a) : ~ is_init a.
Proof.
intro.
assert ((W \₁ RMW) a) by (apply init_w; basic_solver).
unfolder in *; type_solver 20.
Qed.

Lemma no_sb_to_init : sb ≡ sb ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
split; [|basic_solver].
unfold sb; rewrite ext_sb_to_non_init at 1; basic_solver.
Qed.

Lemma no_rf_to_init WF : rf ≡ rf ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
split; [|basic_solver].
rewrite (wf_rfD WF) at 1.
rewrite (wf_rfE WF) at 1.
generalize (read_or_rmw_not_init WF).
basic_solver 42.
Qed.

Lemma no_hb_to_init WF : hb ≡ hb ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
split; [|basic_solver].
unfold hb.
rewrite ct_end.
rewrite (no_rf_to_init WF) at 2.
rewrite no_sb_to_init at 2.
basic_solver 42.
Qed.

Lemma init_same_loc a b 
  (A: is_init a) (B: is_init b) (LOC: loc a = loc b): 
  a = b.
Proof.
destruct a, b; desf.
by cut (x = x0); [ins; subst|simpls].
Qed.

(******************************************************************************)
(** ** More properties *)
(******************************************************************************)

Lemma wf_index_helper WF a b :
    E a /\ E b /\ tid a = tid b /\ same_index a b /\ ~ is_init a -> a = b.
Proof.
intros; desf.
forward apply ((wf_index WF) a b).
unfold same_tid, same_index; basic_solver.
basic_solver 12.
Qed.

Lemma sb_semi_total_l x y z 
  WF (N: ~ is_init x) (NEQ: y <> z) (XY: sb x y) (XZ: sb x z): 
  sb y z \/ sb z y.
Proof.
unfold sb in *; unfolder in *; desf.
cut (ext_sb y z \/ ext_sb z y); [basic_solver 12|].
eapply ext_sb_semi_total_l; eauto.
intro A.
assert (y = z).
eapply wf_index_helper; splits; eauto.
by unfold ext_sb in *; destruct y,z; ins; desf; desf.
by unfold ext_sb in *; destruct y,z; ins; desf; desf.
eauto.
Qed.

Lemma sb_semi_total_r x y z 
  WF (N: ~ is_init z) (NEQ: y <> z) (XY: sb y x) (XZ: sb z x): 
  sb y z \/ sb z y.
Proof.
cut ((sb ∪ sb⁻¹) y z); [basic_solver|].
unfold sb in *; unfolder in *; desf.
destruct (classic (is_init y)).
unfold ext_sb; basic_solver.
cut (ext_sb y z \/ ext_sb z y); [basic_solver|].
eapply ext_sb_semi_total_r; eauto.
intro A.
assert (y = z).
eapply wf_index_helper; splits; eauto.
by unfold ext_sb in *; destruct y,z; ins; desf; desf.
eauto.
Qed.

Lemma sb_tid_init x y (SB : sb x y): tid x = tid y \/ is_init x.
Proof.
generalize ext_sb_tid_init; unfold sb in *.
unfolder in *; basic_solver.
Qed.

Lemma sb_tid_init': sb ≡ sb ∩ same_tid ∪ ⦗is_init⦘ ⨾ sb.
Proof.
split; [|basic_solver].
unfold sb.
rewrite ext_sb_tid_init' at 1.
basic_solver 42.
Qed.

Lemma tid_sb WF: ⦗E⦘ ⨾ same_tid ⨾  ⦗E⦘ ⊆ sb^? ∪ sb⁻¹ ∪ (is_init × is_init).
Proof.
unfold sb.
arewrite (same_tid ⊆ same_tid ∩ same_tid).
rewrite tid_ext_sb at 1; relsf; unionL; try basic_solver 12.
unfolder; ins; desf; try basic_solver 12.
forward apply ((wf_index WF) x y); basic_solver.
Qed.

Lemma tid_n_init_sb WF: ⦗E⦘ ⨾ same_tid ⨾ ⦗set_compl is_init⦘  ⨾  ⦗E⦘ ⊆ sb^? ∪ sb⁻¹.
Proof.
unfold sb.
arewrite (same_tid ⊆ same_tid ∩ same_tid).
rewrite tid_ext_sb at 1; relsf; unionL; try basic_solver 12.
unfolder; ins; desf; try basic_solver 12.
forward apply ((wf_index WF) x y); basic_solver.
Qed.

Lemma init_ninit_sb (WF : Wf) x y (INIT : is_init x) (ININE : E x) (INE : E y)
      (NINIT : ~ is_init y): sb x y.
Proof. 
unfold sb, ext_sb; basic_solver.
Qed.

Lemma same_thread WF x y (X : E x) (Y : E y)
      (NINIT : ~ is_init x) (ST : tid x = tid y):
  sb^? x y \/ sb y x.
Proof.
cut (sb^? y x \/ sb x y); [basic_solver|].
generalize tid_n_init_sb.
unfold same_tid; basic_solver 10.
Qed.

Lemma sb_immediate_adjacent WF:
 ⦗fun a => ~ is_init a⦘ ⨾ immediate sb ≡ ⦗fun a => ~ is_init a⦘ ⨾ (adjacent sb ∩ sb).
Proof.
apply immediate_adjacent.
- unfolder; ins; desf; destruct (classic (x=y)); auto.
  forward (apply (@sb_semi_total_r z y x)); eauto; tauto.
- unfolder; ins; desf; destruct (classic (x=y)); auto.
  forward (apply (@sb_semi_total_l z y x)); eauto; tauto.
- apply sb_trans.
- apply sb_irr.
Qed.

Lemma co_co WF: co ⨾ co ⊆ co.
Proof. apply rewrite_trans, WF. Qed.

(******************************************************************************)
(** ** external-internal restrictions *)
(******************************************************************************)

Definition rfe := rf \ sb.
Definition coe := co \ sb.
Definition fre := fr \ sb.
Definition rfi := rf ∩ sb.
Definition coi := co ∩ sb.
Definition fri := fr ∩ sb.

Lemma ri_union_re r : r ≡ r ∩ sb ∪ r \ sb.
Proof. unfolder; split; ins; desf; tauto. Qed.

Lemma rfi_union_rfe : rf ≡ rfi ∪ rfe.
Proof. apply ri_union_re. Qed.
Lemma coi_union_coe : co ≡ coi ∪ coe.
Proof. apply ri_union_re. Qed.
Lemma fri_union_fre : fr ≡ fri ∪ fre.
Proof. apply ri_union_re. Qed.

Lemma ri_dom r d1 d2 (DOM: r ≡ ⦗d1⦘ ⨾ r ⨾ ⦗d2⦘) : r ∩ sb ⊆ ⦗d1⦘ ⨾ r ∩ sb ⨾ ⦗d2⦘.
Proof. rewrite DOM at 1; basic_solver. Qed.
Lemma re_dom r d1 d2 (DOM: r ≡ ⦗d1⦘ ⨾ r ⨾ ⦗d2⦘) : r \ sb ⊆ ⦗d1⦘ ⨾ (r \ sb) ⨾ ⦗d2⦘.
Proof. rewrite DOM at 1; basic_solver. Qed.

Lemma wf_rfiE WF: rfi ≡ ⦗E⦘ ⨾ rfi ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_rfE WF)). Qed.
Lemma wf_coiE WF: coi ≡ ⦗E⦘ ⨾ coi ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_coE WF)). Qed.
Lemma wf_friE WF: fri ≡ ⦗E⦘ ⨾ fri ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_frE WF)). Qed.
Lemma wf_rfeE WF: rfe ≡ ⦗E⦘ ⨾ rfe ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_rfE WF)). Qed.
Lemma wf_coeE WF: coe ≡ ⦗E⦘ ⨾ coe ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_coE WF)). Qed.
Lemma wf_freE WF: fre ≡ ⦗E⦘ ⨾ fre ⨾ ⦗E⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_frE WF)). Qed.
Lemma wf_rfiD WF : rfi ≡ ⦗W⦘ ⨾ rfi ⨾ ⦗R⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_rfD WF)). Qed.
Lemma wf_coiD WF : coi ≡ ⦗W⦘ ⨾ coi ⨾ ⦗W⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_coD WF)). Qed.
Lemma wf_friD WF : fri ≡ ⦗R⦘ ⨾ fri ⨾ ⦗W⦘.
Proof. split; [|basic_solver]. apply (ri_dom (wf_frD WF)). Qed.
Lemma wf_rfeD WF : rfe ≡ ⦗W⦘ ⨾ rfe ⨾ ⦗R⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_rfD WF)). Qed.
Lemma wf_coeD WF : coe ≡ ⦗W⦘ ⨾ coe ⨾ ⦗W⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_coD WF)). Qed.
Lemma wf_freD WF : fre ≡ ⦗R⦘ ⨾ fre ⨾ ⦗W⦘.
Proof. split; [|basic_solver]. apply (re_dom (wf_frD WF)). Qed.

(******************************************************************************)
(** ** properties of external/internal relations *)
(******************************************************************************)

Lemma seq_ii r1 r2 r3 (A: r1 ⨾ r2 ⊆ r3): r1 ∩ sb ⨾ r2 ∩ sb ⊆ r3 ∩ sb.
Proof.
generalize sb_trans.
unfolder in *; basic_solver 21.
Qed.

Lemma re_ri WF  r r' (IRR: irreflexive r)  (IRR2: irreflexive (r ⨾ sb))
  (N: r ⊆ r ⨾  ⦗ fun x => ~ is_init x ⦘): (r \ sb) ⨾ (r' ∩ sb) ⊆ r ⨾  r' \ sb.
Proof.
rewrite N at 1.
unfolder; ins; desf; splits; eauto.
intro.
eapply sb_semi_total_r with (x:=y) (y:=x) in H1; eauto.
by desf; revert IRR2; basic_solver.
eby intro; subst; eapply IRR.
Qed.

Lemma ri_re WF  r r' (IRR: irreflexive r')  (IRR2: irreflexive (r' ⨾ sb)): 
 ⦗ fun x => ~ is_init x ⦘ ⨾ (r ∩ sb) ⨾ (r' \ sb) ⊆ r ⨾  r' \ sb.
Proof.
unfolder; ins; desf; splits; eauto.
intro.
eapply sb_semi_total_l with (x:=x) (y:=z) (z:=y) in H4; eauto.
by desf; revert IRR2; basic_solver.
eby intro; subst; eapply IRR.
Qed.

Lemma rfi_in_sbloc WF : rf ∩ sb ⊆ restr_eq_rel loc sb.
Proof. rewrite wf_rfl; basic_solver 12. Qed.
Lemma coi_in_sbloc WF : co ∩ sb ⊆ restr_eq_rel loc sb.
Proof. rewrite wf_col; basic_solver 12. Qed.
Lemma fri_in_sbloc WF : fr ∩ sb ⊆ restr_eq_rel loc sb.
Proof. rewrite (funeq_same_loc (loceq_fr WF)).
unfolder; unfold Events.same_loc in *.
ins; desf; splits; eauto; congruence.
Qed.
Lemma rfi_in_sbloc' WF : rfi ⊆ sb ∩ same_loc.
Proof. generalize (wf_rfl WF); unfold rfi; basic_solver 12. Qed.
Lemma coi_in_sbloc' WF : coi ⊆ sb ∩ same_loc.
Proof. generalize (wf_col WF); unfold coi; basic_solver 12. Qed.
Lemma fri_in_sbloc' WF : fri ⊆ sb ∩ same_loc.
Proof. generalize (wf_frl WF); unfold fri; basic_solver 12. Qed.

End Execution.

(******************************************************************************)
(** ** Tactics *)
(******************************************************************************)

#[export] Hint Unfold rfe coe fre rfi coi fri : ie_unfolderDb.
Tactic Notation "ie_unfolder" :=  repeat autounfold with ie_unfolderDb in *.


