(******************************************************************************)
(** * Subsequences: definitions and lemmas        *)
(** * Two alternative presentations:              *)
(** *  1. a recursive definiton                   *)
(** *  2. a witness definirion: there exists a monontone mapping         *)
(******************************************************************************)

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.
Require Import List.
Require Import Basic.

Set Implicit Arguments.

Section Subseq.

Variable A : Type.

Inductive subseq : list A -> list A -> Prop :=
| sub_nil (l : list A) : subseq nil l
| sub_fst (a : A) (l1 l2 : list A) 
          (SUB: subseq l1 l2) : subseq  (a :: l1) (a :: l2)
| sub_nfst (a : A) (l1 l2 : list A) 
          (SUB: subseq l1 l2) : subseq  l1 (a :: l2).

Lemma subseq_reflexive (L : list A): 
  subseq L L.
Proof.
  induction L. econs. econs. auto.
Qed.

Lemma subseq_length (l1 l2 : list A) (SUB: subseq l1 l2) :
  length l1 <= length l2.
Proof.
induction SUB; ins; lia.
Qed.

Lemma subseq_In (l1 l2 : list A) (SUB: subseq l1 l2) a:
  In a l1 -> In a l2.
Proof.
induction SUB.
desf.
simpl; ins; desf; eauto.
simpl; ins; desf; eauto.
Qed.

Definition subseq_witness {A: Type} (l1 l2 : list A) (f : nat -> nat) := 
  << MON: forall n1 n2, n1 < n2 -> f n1 < f n2 >> /\
  << NTH: forall n, nth_error l1 n = nth_error l2 (f n) >> /\
  << LEN: forall n, n < length l1 -> f n < length l2 >>.

Lemma subseq_mapping (l1 l2 : list A) (SUB: subseq l1 l2) :
  exists (f : nat -> nat), subseq_witness l1 l2 f.
Proof.
induction SUB; ins; desc.
- exists (fun n => n + length l); red; splits; eauto.
  + ins; lia.
  + ins.
    rewrite nth_error_nil.
    symmetry.
    apply nth_error_None.
    lia.
  + ins; lia.
- exists (fun n => match n with
                   | 0 => 0
                   | S n' => S (f n')
                   end).
  red in IHSUB; desf.
  red; splits; ins; desf; try lia.
  + apply -> Nat.succ_lt_mono. apply MON. lia.
  + eauto.
  + apply -> Nat.succ_lt_mono. apply LEN. lia.
- exists (fun n => S (f n)).
  red in IHSUB; desf.
  red; splits; ins; desf; try lia.
  + apply -> Nat.succ_lt_mono. apply MON. lia.
  + apply -> Nat.succ_lt_mono. apply LEN. lia.
Qed.

Lemma subseq_mapping_conv (l1 l2 : list A) 
  (f : nat -> nat)
  (MON: forall n1 n2 (LEN: n2 < length l1), n1 < n2 -> f n1 < f n2)
  (NTH: forall n, n < length l1 -> nth_error l1 n = nth_error l2 (f n)):
  subseq l1 l2.
Proof.
generalize dependent f.
generalize dependent l1.
induction l2; ins.
- destruct l1; [by econs|].
specialize (NTH 0 (Nat.lt_0_succ _)).
destruct (f 0); ins.
- ins; destruct (f 0) eqn:F.
+ ins; destruct l1.
by econs.
assert (NTH':=NTH).
specialize (NTH 0 (Nat.lt_0_succ _)); ins; rewrite F in NTH; ins; desf.
apply sub_fst.
eapply IHl2 with (f:= fun n => Nat.pred (f (S n))); ins.
* apply (Nat.pred_lt_mono (f (S n1)) (f (S n2))).
-- intro H0; specialize (MON 0 (S n1)).
   rewrite H0 in MON; eapply Nat.nlt_0_r; eapply MON; lia.
-- apply MON; lia.
* specialize (NTH' (S n)); ins.
rewrite NTH'; try lia.
destruct (f (S n)) eqn:FS; ins; try lia.
specialize (MON 0 (S n)). 
   rewrite FS in MON.
exfalso; eapply Nat.nlt_0_r; eapply MON; lia.
+ ins; destruct l1.
by econs.
apply sub_nfst.
eapply IHl2 with (f:= fun n => Nat.pred (f n)); ins.
* apply (Nat.pred_lt_mono (f n1) (f n2)); eauto.
intro; specialize (MON 0 n1).
destruct (f n1) eqn:FS; ins.
exfalso.
destruct n1; desf.
congruence.
eapply Nat.nlt_0_r; eapply MON; lia.
* specialize (NTH n0); ins.
rewrite NTH; try lia.
destruct (f n0) eqn:FS; ins.
exfalso.
destruct n0; desf.
congruence.
specialize (MON 0 (S n0)).
rewrite FS in MON.
eapply Nat.nlt_0_r; eapply MON; lia.
Qed.

Lemma subseq_witness_nth 
  (l1 l2 : list A) (f : nat -> nat) 
  (SUB: subseq_witness l1 l2 f) : 
  forall n d1 d2, n < length l1 -> nth n l1 d1 = nth (f n) l2 d2.
Proof.
red in SUB; desc.
ins.
apply nth_error_nth.
rewrite NTH; try done.
apply nth_error_nth'; eauto.
Qed.

End Subseq.


