(******************************************************************************)
(** * Basic definitions for handling "write lists"                       *)
(** * useful for representing lists as functions from Nat + list size    *)
(******************************************************************************)

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Set Implicit Arguments.

Section WL.

Variable A : Type.

Definition in_wl (wl : nat -> A) N e : Prop :=
  exists n, n < N /\ wl n = e.

Definition order_wl (wl : nat -> A) N e1 e2 : Prop :=
  exists n1 n2, n1 < n2 /\ n2 < N /\
                wl n1 = e1 /\ wl n2 = e2. 

Definition extend_wl (wl : nat -> A) (e : A) :=
 fun n => if (Nat.eqb 0 n) then e else wl (pred n).

Lemma extend_wl_wl wl :
  extend_wl (fun n => wl (S n)) (wl 0) = wl.
Proof.
apply functional_extensionality; ins.
unfold extend_wl; desf; desf.
rewrite Nat.succ_pred; auto.
Qed.

Lemma extend_wl_in_wl wl e n :
  in_wl (extend_wl wl e) (S n) e.
Proof.
red; exists 0; splits; auto; lia.
Qed.

Lemma in_wl_monotone wl n :
  in_wl (fun m => wl (S m)) n ⊆₁ in_wl wl (S n).
Proof.
unfold in_wl; unfolder; ins; desf.
exists (S n0); splits; try done.
lia.
Qed.

Lemma order_wl_monotone wl n :
  order_wl (fun m => wl (S m)) n ⊆ order_wl wl (S n).
Proof.
unfold order_wl; unfolder; ins; desf.
exists (S n1), (S n2); splits; try done; lia.
Qed.

Lemma order_wlE wl N :
  order_wl wl N ≡
  <| in_wl wl N |> ;; order_wl wl N ;; <| in_wl wl N |>.
Proof.
unfold order_wl, in_wl; split; unfolder; ins; desc; subst; eauto 8.
assert (n1 < N) by lia.
splits; desf; eauto 8.
Qed.

Lemma in_wl_extend_wl wl e n :
  in_wl (extend_wl wl e) (S n) ≡₁ in_wl wl n ∪₁ (eq e).
Proof.
unfold extend_wl, in_wl.
unfolder; split; ins; desf; eauto.
- left; exists n1; splits; ins; lia.
- exists (S n0); splits; [lia| eauto].
- exists 0; splits; [lia| eauto].
Qed.

Lemma order_wl_extend_wl wl e n :
  order_wl (extend_wl wl e) (S n) ≡ 
  order_wl wl n 
  ∪ (eq e) × (in_wl wl n).
Proof.
unfold order_wl, extend_wl, in_wl.
unfolder; split; ins; desf; simpl; eauto; try lia.
- apply Nat.succ_lt_mono in H0. eauto 10.
- apply Nat.succ_lt_mono in H0, H. eauto 10.
- exists (S n1), (S n2).
rewrite <- !pred_Sn.
splits; eauto; lia.
- exists 0, (S n0); splits; eauto.
lia.
lia.
Qed.

Definition worder_wl (wl : nat -> A) N e1 e2 : Prop :=
  exists n1 n2, n1 <= n2 /\ n2 < N /\
                wl n1 = e1 /\ wl n2 = e2. 

Lemma worder_wl_monotone wl n :
  worder_wl (fun m => wl (S m)) n ⊆ worder_wl wl (S n).
Proof.
unfold worder_wl; unfolder; ins; desf.
exists (S n1), (S n2); splits; try done; lia.
Qed.

Lemma worder_wlE wl N :
  worder_wl wl N ≡
  <| in_wl wl N |> ;; worder_wl wl N ;; <| in_wl wl N |>.
Proof.
unfold worder_wl, in_wl; split; unfolder; ins; desc; subst; eauto 8.
assert (n1 < N) by lia.
splits; desf; eauto 8.
Qed.

Lemma worder_wl_extend_wl wl e n :
  worder_wl (extend_wl wl e) (S n) ≡ 
  worder_wl wl n 
  ∪ (eq e) × ((in_wl wl n) ∪₁ (eq e)).
Proof.
unfold worder_wl, extend_wl, in_wl.
unfolder; split; ins; desf; simpl; eauto; try lia.
- apply Nat.succ_lt_mono in H0. eauto 10.
- left. eexists n2,n0; splits; eauto; lia.
- exists (S n1), (S n2); splits; eauto; lia.
- exists 0, (S n0); splits; eauto; lia.
- exists 0, 0; splits; eauto; lia.
Qed.

End WL.

Require Import Labels Events.

Section WLn.

Definition event_in_wl (wl : nat -> Loc -> Event) N :=
 fun e => exists ws x, (in_wl wl N) ws /\ e = ws x.

Lemma event_in_wl_k wl N k (Z : k < N) x:
  event_in_wl wl N (wl k x).
Proof.
 red. eexists _,_; splits; eauto.
 red; eexists; splits; eauto.
Qed.

Definition event_order_in_wl (wl : nat -> Loc -> Event) N :=
 fun e1 e2 => exists ws1 ws2 x y, 
   (worder_wl wl N) ws1 ws2 /\ e1 = ws1 x /\ e2 = ws2 y.

Lemma event_order_in_wl_k wl N k1 k2 (Z : k2 < N) (L : k1 <= k2) x y:
event_order_in_wl wl N  (wl k1 x) (wl k2 y).
Proof.
 red. eexists _,_,_,_; splits; eauto.
 red; eexists _,_; splits; eauto.
Qed.

Lemma event_order_wlE wl N :
  event_order_in_wl wl N ≡
  <| event_in_wl wl N |> ;; event_order_in_wl wl N ;; <| event_in_wl wl N |>.
Proof.
unfold event_order_in_wl, event_in_wl, worder_wl, in_wl.
split; unfolder; ins; desf; subst; eauto 16.
assert (n1 < N) by lia.
all: splits; desf; eauto 16.
Qed.

Lemma event_order_reflexive wl N e (IN : event_in_wl wl N e):
  event_order_in_wl wl N e e.
Proof.
unfold event_order_in_wl, event_in_wl, in_wl in *; desf.
exists (wl n), (wl n), x, x; splits; eauto.
red; eauto 10.
Qed.

Lemma event_order_zero wl N e (IN : event_in_wl wl N e) x:
event_order_in_wl wl N (wl 0 x) e.
Proof.
unfold event_order_in_wl, event_in_wl, in_wl in *; desf.
exists (wl 0), (wl n), x, x0; splits; eauto.
exists 0, n; splits; eauto 10. lia.
Qed.

Lemma event_in_wl_extend_wl wl ws n (Z : n > 0):
  event_in_wl (extend_wl wl ws) (S n) ≡₁ 
  event_in_wl wl n ∪₁ (fun e => exists x, ws x = e).
Proof.
unfold event_in_wl.
split; unfolder; ins; desf; eauto.
- apply in_wl_extend_wl in H.
  unfolder in H; desf; eauto.
- eexists _, _.
splits; eauto.
apply in_wl_extend_wl.
basic_solver.
- eexists _, _.
splits; eauto.
apply in_wl_extend_wl.
basic_solver.
Qed.

Lemma event_order_in_wl_extend_wl wl ws n (Z : n > 0) :
  event_order_in_wl (extend_wl wl ws) (S n) ≡ 
  event_order_in_wl wl n 
  ∪ ((fun e => exists x, ws x = e) × (event_in_wl wl n))
  ∪ ((fun e => exists x, ws x = e) × (fun e => exists x, ws x = e)).
Proof.
unfold event_order_in_wl.
split; unfolder; ins; desf; eauto.
- apply worder_wl_extend_wl in H.
  unfold event_in_wl.
  unfolder in H; desf; eauto 22.
- eexists _, _,_,_.
splits; eauto.
apply worder_wl_extend_wl.
basic_solver.
- unfold event_in_wl in *; desf.
eexists _, _,_,_.
splits; eauto.
apply worder_wl_extend_wl.
basic_solver.
- 
eexists _, _,_,_.
splits; eauto.
apply worder_wl_extend_wl.
basic_solver.
Qed. 

End WLn.

